#!/bin/bash
#shellcheck disable=SC2120 disable=SC2119

shellfu import arr
shellfu import hackr
shellfu import jat

#
# Help selectively apply hacks and workarounds
#
# This module aims to make it easier to apply workarounds and hacks
# to system in a transparent and maintainable way.  Each hack or
# workaround is implemented as a simple module of few functions
# and applied using interface provided by this module.  This module
# then takes care of selection, logging and keeps state as of which
# hacks are applied.
#
#
# =head2 IMPLEMENTING HACKS
#
# Each hack must have a unique identifier consisting of of alphanumeric
# symbols and underscore (e.g. `bz1234` is a valid hack). (IOW, must match
# `[[:alnum:]_]+` regular expression.)
#
# A hack is implemented as a "driver", ie. a local function or a code
# snippet that implements simple interface consisting of a single argument:
# "action" to be handled.  There are five actions:
#
#  *  'applies' must return true if the particular hack should be applied
#      under current circumstances (eg. `dist__test id eq rhel ver.x eq 5`),
#
#  *  'apply' should apply the actual hack,
#
#  *  'remove' to revert effect of 'apply',
#
#  *  'tags' to print all tags the hack should be selectable by, one
#     per line,
#
#  *  and 'justify' to provide hack justification (see JUSTIFYING HACKS
#     section).
#
# If driver is implemented as a local function, it must be called
# hack__drv_ID(), where ID is the identifier mentioned
# above.
#
# In case of Beakerlib snippet, the driver implementation must be saved
# as a file 'hack.sh' under a directory named same as hack ID, under one
# of directories listed in $HACK__PATH variable.  For example,
# if value of $HACK__PATH is /foo:/bar, /foo/myhack/hack.sh
# is a valid snippet constituting driver for hack identified as `myhack`,
# and `hack__apply myhack` is a valid call.
#
# Note: this module contains declaration of hack__drv_SKEL()
# function; feel free to copy it as a template.
#
#
# =head3 File attachments
#
# Hacks that are implemented as Beakerlib snippets will be executed in
# their directory; that means you can store additional files in that
# directory (along with hack.sh) and the hack code can access them
# directly.
#
# For example, given following tree:
#
#    └── hacks
#        ├── bz1234
#        │   └── hack.sh
#        ├── bz5678
#        │   ├── file.gz
#        │   └── hack.sh
#        └── bz90ab
#            └── hack.sh
#
# hack 'bz5678' can include code that works with 'file.gz' without need
# of further path specification.
#
#
# =head3 Internal hack API
#
# Several convenience functions are available from within hack; see
# documentation of distribution/hackr.
#
#
# =head2 EXAMPLE
#
# Following is a minimal example of a test with a working hack implemented as
# a Bash function:
#
#     hack__drv_foo() {
#         local action=$1
#         case $action in
#             applies)
#                 dist__test id eq rhel  ver.x le 5   || return 1
#                 jat__cmd test -f /etc/bar/some_file || return 1
#                 return 0
#                 ;;
#             apply)
#                 jat__eval "echo foo >/etc/bar/some_file"
#                 ;;
#             remove)
#                 jat__cmd rm /etc/bar/some_file
#                 ;;
#             tags)
#                 echo suite::bar
#                 ;;
#             justify)
#                 echo Hey, I'm Mr. Meeseeks, look at me!
#                 echo http://bugzilla.redhat.com/show_bug.cgi?id=123456
#                 ;;
#         esac
#     }
#
#     shellfu import hack
#
#     hack__apply suite::bar
#
#     jat__pstartt
#         # the actual test
#     jat__pend
#
#     hack__remove suite::bar
#     hack__report
#
# The above test will run with the 'foo' hack applied and clean up
# afterwards, submitting the fact as a YAML report.
#
#
# =head2 JUSTIFYING HACKS
#
# Hacks are bad, because they bring risk of making the code counter-intuitive,
# expensive, hard to maintain, broken, incorrect, or all of the above.  But The
# Blade of Reality can be sharp and brutal: sometimes hacks are too hard
# to avoid, at least the temporary ones.  So if we have to have them, let's
# prevent them from cumulating, shall we?
#
# Hence, each hack must provide justification for its existence.  (If that
# sentence feels harsh, try to think of hacks like Meeseeks: they exist just
# for one purpose, they desperately want to die but can't before they finish
# their job.)
#
# An acceptable justification:
#
#  *  must explain "why/how" of the hack, including details of its anticipated
#     effect,
#
#  *  must enable reader to assess whether the hack is *still* necessary,
#
#  *  can be in form of one or more URIs.
#
# Usually, the easiest (and often the best) way to provide both is to have
# a Bugzilla/RT/Jira/Github/Gitlab/... ticket and just provide URL.
# Alternatively (eg. in emergency cases), link to less formal resource such
# as mailing list archive can be acceptable.  Writing the justification in
# human-readable text is actually least preferred option because it may require
# maintenance (by people who know about the domain) and is unlikely to provide
# the second requirement.
#
#
# =head2 TAGS
#
# That is, each TAG has form of `CATEGORY::VALUE`.   To this day, we only
# define one CATEGORY, though:
#
#  *  *suite* test suite name, by which the hack is required.  This is normally
#     the same as component name, also known as "srpm name", and sometimes
#     referred to using ambiguous term "package".
#
#     Note that a common case where *suite* is different from the actual
#     component being tested is AVC workaround.
#
# This tag mechanism is inspired by debtags, which is an implementation of
# a "Faceted Classification".
#

#
# List of paths where to look for hacks
#
# Similar to the standard PATH environment variable, this
# variable contains colon-separated list of paths to consider
# when looking for hack implementation.
#
# Note that a hack implemented as function always has higher priority
# than any of these paths.  See section IMPLEMENTING HACKS of this manual
# for details.
#
# Default value is path to 'hacks' sub-directory of directory where
# library(distribution/hack) is installed.
#
HACK__PATH=${HACK__PATH:-}

#
# Allow hacks to be applied?
#
# Can be 'allow' to allow, 'deny' to deny all hacks and 'skip'
# to allow but skip all of them.
#
HACK__POLICY=${HACK__POLICY:-allow}

#
# Valid exit status expression
#
# Comma-separated list of exit statuses or exit status ranges,
# defining valid exit statuses for hack handler.
#
# Exit statuses that don't match this expression will be reported
# as error.
#
# Default value is '0,1', ie. only zero and one are accepted.
#
HACK__VALID_ES=${HACK__VALID_ES:-0,1}

hack__apply() {
    #
    # Apply hacks $@
    #
    # Usage:
    #     hack__apply TAG|HACK..
    #
    # If TAG is passed, it stands for selection of all hacks found at all
    # known places ($HACK__PATH, or defined as driver functions)
    # that include this tag.  See TAGS section of this manual.
    #
    # HACK is merely a hack name.
    #
    # Note that it's almost always preferred to use TAGs and maintain hacks by
    # adding them to a well-known path and selecting them by tags.  That way
    # you can add/remove them without need to keep list of hack names.
    #
    local args=("$@")
    local hacks=()
    local do_phase=true
    while true; do case $1 in
        "")     break ;;
        *::*)   readarray -t -O ${#hacks[@]} hacks <<<"$(__hack__select "$1")"; shift ;;
        *)      hacks+=("$1"); shift ;;
    esac done
    test -n "${hacks[*]}" \
     || jat__log_error "usage: hack__apply TAG|HACK.."
    jat__meta ptype >&/dev/null && do_phase=false
    $do_phase && jat__pstarts -h "consider and apply hacks: ${args[*]}"
        case $HACK__POLICY in
            allow|skip)
                for _hackr__hname in "${hacks[@]}"; do
                    __hack__apply1
                done
                ;;
            deny)
                jat__log_info "hacks are disabled by policy: HACK__POLICY=$HACK__POLICY"
                ;;
        esac
    $do_phase && jat__pend
    true
}

hack__find() {
    #
    # Print list of all hacks with state $1 (or any hacks seen)
    #
    # Usage:
    #     hack__find [STATE]
    #
    # Without any arguments, all hacks seen by hack__apply()
    # are listed (ie. including hacks that were skipped or failed to
    # apply).
    #
    # If STATE is provided, only hacks with a matching state are listed.
    #
    # Exits with zero if any hacks were found, one if no hacks were
    # found and two or above in case of error.
    #
    local state_re=${1:-.*}
    local path
    local buff
    find "$__hack__dir/hacks" -maxdepth 2 -mindepth 2 -name state \
      | while read -r path; do
            grep -qx "$state_re" "$path" || continue
            buff=${path#$__hack__dir/hacks/}
            echo "${buff%/state}"
        done \
      | grep .
}

hack__remove() {
    #
    # Remove hacks $@
    #
    # Usage:
    #
    #     hack__remove TAG|HACK..
    #
    # Remove specified hacks.  Same syntax as hack__apply()
    #
    local args=("$@")
    local hacks=()
    local do_phase=true
    while true; do case $1 in
        "")     break ;;
        *::*)   readarray -t -O ${#hacks[@]} hacks <<<"$(__hack__select "$1")"; shift ;;
        *)      hacks+=("$1"); shift ;;
    esac done
    test -n "${hacks[*]}" \
     || jat__log_error "usage: hack__remove TAG|HACK.."
    jat__meta ptype >&/dev/null && do_phase=false
    $do_phase && jat__pstartc -h "remove hacks: ${args[*]}"
        for _hackr__hname in "${hacks[@]}"; do
            __hack__remove1
        done
    $do_phase && jat__pend
    true
}

hack__report() {
    #
    # Create hack state report YAML and jat__submit it
    #
    local reptmp
    local do_phase=true
    jat__meta ptype >&/dev/null && do_phase=false
    $do_phase && jat__pstartc -h "report hack state"
        reptmp=$(mktemp -t hack__report.XXXXXXXX)
        __hack__report >"$reptmp"
        jat__submit "$reptmp" "hack-report.yaml"
        rm "$reptmp"
    $do_phase && jat__pend
    true
}

#           #                                                            #
# TEMPLATES # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#           #                                                            #

hack__drv_SKEL() {
    #
    # Skeleton for hack driver
    #
    # Use this as a starting point for implementating own hack.
    # See IMPLEMENTING HACKS section of manual.
    #
    local action=$1     # action to handle
    case $action in
        justify)    echo "https://bugzilla.example.com/show_bug.cgi?id=123456" ;;
        applies)    false ;;
        apply)      jat__cmd -h "apply hack $(hackr__name)" true ;;
        tags)       echo suite::REPLACE_ME ;;
        remove)     jat__cmd -h "remove hack $(hackr__name)" true ;;
    esac
}


#          #                      please do not apply hack to your brain #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                           that would let you past this line #


#
# Path to storage directory root
#
# This is where data like hack states are kept
#
__hack__dir=${__hack__dir:-/var/tmp/hack}

__hack__report() {
    #
    # Create report of all hacks
    #
    local l_time
    local l_test
    local l_action
    local l_es
    echo "---"
    if hack__find >/dev/null; then
        echo "hacks:"
        #shellcheck disable=SC2045
        for _hackr__hname in $(hack__find); do
            echo "    -"
            echo "        id: $_hackr__hname"
            echo "        state: $(__hack__rkey state)"
            echo "        justification: |-4"
            __hack__rkey justification \
              | sed 's/^/            /'
            echo "        code: |-4"
            __hack__rkey code \
              | sed 's/^/            /' \
              | sed 's/^ *$//'
            echo "        source: $(__hack__rkey source)"
            echo "        log:"
            __hack__rkey log \
              | while read -r l_time l_test l_action l_es; do
                    echo "            -"
                    echo "                time: $l_time"
                    echo "                test: $l_test"
                    echo "                action: $l_action"
                    echo "                exit_status: $l_es"
                done
        done
    else
        echo "hacks: []"
    fi
}

__hack__apply1() {
    #
    # Apply hack $_hackr__hname if needed
    #
    __hack__isfun || __hack__iscode || {
        jat__log_warning "skipping unknown hack: missing driver: $_hackr__hname"
        return 3
    }
    __hack__wkey state new
    __hack__identify
    jat__log_info "considering hack: $_hackr__hname"
    __hack__justify
    test "$HACK__POLICY" == "skip" && {
        jat__log_info "skipping hack due to policy: HACK__POLICY='$HACK__POLICY'"
        __hack__wkey state skipped
        return 1
    }
    __hack__handle applies || {
        jat__log_info "$_hackr__hname: does not apply"
        __hack__wkey state skipped
        return 1
    }
    if __hack__handle apply; then
        __hack__wkey state applied
    else
        __hack__wkey state apply_failed
        jat__log_error "hack application failed: $_hackr__hname"
    fi
}

__hack__es_match() {
    #
    # True if exit status $2 matches expression $1
    #
    local expr=$1   # ES_EXPR expression
    local es=$2     # compared exit status
    local part      # each part of ES_EXPR
    for part in ${expr//,/ }; do
        test -n "$part" || continue
        #FIXME: a rather funny implementation (works, though...)
        eval "echo {${part/-/..}}" | grep -qwFe "$es" && return 0
    done
    return 1
}

__hack__identify() {
    #
    # Get and report hack identity
    #
    local src   # source info (fn name or snippet path)
    if __hack__isfun; then
        src="hack__drv_$_hackr__hname()"
        __hack__wkey source "$src"
        __hack__wkey code "$(
            type "hack__drv_$_hackr__hname" \
              | tail -n +4 \
              | head -n -1
        )"
        #FIXME: we should use `declare -fp` but it does not find the function
    else
        src=$(__hack__shfile)
        __hack__wkey source "$src"
        __hack__wkey code "$(<"$src")"
    fi
}

__hack__justify() {
    #
    # Get, report and jat__log_info justification text
    #
    local text
    local line
    text=$(__hack__handle justify)
    test -n "$text" \
     || jat__log_warning "missing justification: $_hackr__hname"
    jat__log_info "$_hackr__hname: justification:"
    while read -r line; do
        jat__log_info "$_hackr__hname:    $line"
    done <<<"$text"
    __hack__wkey justification "$text"
}

__hack__remove1() {
    #
    # Apply hack $_hackr__hname if needed
    #
    local hstate
    hstate=$(__hack__state) || return 3
    case $hstate in
        applied)
            :
            ;;
        apply_failed)
            jat__log_info "skipping removal after failed applying: $_hackr__hname"
            return 2
            ;;
        removed)
            jat__log_warning "skipping double removal: $_hackr__hname"
            return 3
            ;;
        skipped)
            jat__log_info "nothing to remove; hack skipped application: $_hackr__hname"
            return 1
            ;;
    esac
    __hack__isfun || __hack__iscode || {
        jat__log_warning "skipping unknown hack: missing driver: $_hackr__hname"
        return 3
    }
    if __hack__handle remove; then
        __hack__wkey state removed
    else
        __hack__wkey state remove_failed
        jat__log_error "hack removal failed: $_hackr__hname"
    fi
}

__hack__wkey() {
    #
    # Write value $2 as key $1 of hack $_hackr__hname
    #
    local key=$1
    local value=$2
    mkdir -p "$__hack__dir/hacks/$_hackr__hname"
    echo "$value" > "$__hack__dir/hacks/$_hackr__hname/$key"
}

__hack__akey() {
    #
    # Append value $2 as key $1 of hack $_hackr__hname
    #
    local key=$1
    local value=$2
    mkdir -p "$__hack__dir/hacks/$_hackr__hname"
    echo "$value" >> "$__hack__dir/hacks/$_hackr__hname/$key"
}

__hack__rkey() {
    #
    # Read value of key $1 of hack $_hackr__hname
    #
    local key=$1
    cat "$__hack__dir/hacks/$_hackr__hname/$key"
}

__hack__state() {
    #
    # Read state of hack $_hackr__hname
    #
    local sbody
    local sfile="$__hack__dir/hacks/$_hackr__hname/state"
    sbody=$(<"$sfile") || {
        jat__log_error "error reading hack state: $sfile"
        return 3
    }
    case $sbody in
        skipped|applied|removed|apply_failed|remove_failed) : ;;
        *) jat__log_error "invalid hack state: '$sbody' in $sfile"
           return 3 ;;
    esac
    echo "$sbody"
}

__hack__handle() {
    #
    # Call action $1 on hack $_hackr__hname driver
    #
    local _hackr__aname=$1
    local _hackr__shpath
    local _hackr__es
    debug -v _hackr__aname _hackr__hname
    if __hack__isfun; then
        (
            "hack__drv_$_hackr__hname" "$_hackr__aname"
        ); _hackr__es=$?
    else
        _hackr__shpath=$(__hack__shfile)
        (
            hack__drv_ADHOC() {
                #shellcheck disable=SC1090
                . "$_hackr__shpath"
            }
            cd "${_hackr__shpath%/*}" || {
                jat__log_error "could not chdir to hack home"
                return 4
            }
            hack__drv_ADHOC "$_hackr__aname"
        ); _hackr__es=$?
    fi
    __hack__es_match "$HACK__VALID_ES" "$_hackr__es" \
     || jat__log_error \
            "invalid exit status: $_hackr__es (not $HACK__VALID_ES)" \
            "(hint: set HACK__VALID_ES to avoid this error)"

    __hack__akey log "$(date -Ins) $TEST $_hackr__aname $_hackr__es"
    return "$_hackr__es"
}

__hack__isfun() {
    #
    # true if hack $_hackr__hname is function
    #
    type -t "hack__drv_$_hackr__hname" >/dev/null
}

__hack__iscode() {
    #
    # true if hack $_hackr__hname is shell code
    #
    __hack__shfile >/dev/null
}

__hack__shfile() {
    #
    # Print path to shell code file for hack $_hackr__hname
    #
    # True if a file was found, false otherwise
    #
    local dir   # a directory in $HACK__PATH
    case $_hackr__hname in
        */*)
            if test -f "$_hackr__hname/hack.sh"; then
                echo "$_hackr__hname/hack.sh"
                return 0
            else
                jat__log_warning "no such file: $_hackr__hname/hack.sh"
                return 3
            fi
            ;;
    esac
    for dir in $(__hack__pathelems); do
        test -f "$dir/$_hackr__hname/hack.sh" \
         && echo "$dir/$_hackr__hname/hack.sh" \
         && return 0
    done
    return 1
}

__hack__pathelems() {
    #
    # List valid elements of $HACK__PATH
    #
    local dir   # a directory in $HACK__PATH
    local path=$HACK__PATH
    #shellcheck disable=SC2154
    test -n "$path" || path="$hack__LibraryDir/hacks"
    for dir in ${path//:/ }; do
        test -n "$dir" || continue
        test -d "$dir" || continue
        echo "$dir"
    done
}

__hack__select() {
    #
    # Select hacks based on tags query
    #
    local tag
    local map
    map=$(__hack__map | cut -d' ' -f1,2)
    for tag in "$@"; do
        map=$(grep -w "$tag" <<<"$map")
    done
    grep -o '^[^ ]\+' <<<"$map"
}

__hack__map() {
    #
    # Create hack map
    #
    # Print list of hack metadata (name, tags and implementation).  One line per
    # found hack, format of each line being:
    #
    #     HNAME HTAGS HIMPL
    #
    # where HNAME is hack name, HTAGS is comma-separated list of tags, HIMPL is
    # driver implementation (function name or path to the driver) and items are
    # delimited by single space.
    #
    local _hackr__hdir
    local _hackr__hname
    local _hackr__hpath
    local _hackr__fname
    local _hackr__htags
    for _hackr__hdir in $(__hack__pathelems); do
        find "$_hackr__hdir" -maxdepth 2 -mindepth 2 -type f -name hack.sh \
          | grep -vE -e '/SKEL/hack[.]sh$' -e '/SKEL[.].*$' \
          | sed 's:/hack.sh::' \
          | while read -r _hackr__hpath; do
                _hackr__hname=${_hackr__hpath##*/}
                _hackr__htags=$(
                    hack__drv_ADHOC() {
                        #shellcheck disable=SC1090,SC1091
                        . "$_hackr__hpath/hack.sh"
                    }
                    cd "$_hackr__hpath" || {
                        jat__log_error "could not chdir to hack home"
                        return 4
                    }
                    hack__drv_ADHOC tags | tr ' ' '\n' | paste -sd,
                )
                echo "$_hackr__hname $_hackr__htags $_hackr__hpath/hack.sh"
            done
    done
    for _hackr__fname in $(set | grep '^hack__drv_.* \(\)$' | grep -v -e 'SKEL' -e 'SKEL[.].*' | cut -d' ' -f1); do
        _hackr__hname=${_hackr__fname#hack__drv_}
        _hackr__htags=$(
            "$_hackr__fname" tags | tr ' ' '\n' | paste -sd,
        )
        echo "$_hackr__hname $_hackr__htags $_hackr__fname()"
    done
}

#shellfu module-version=0.0.14
