#!/bin/bash

shellfu import jat


#
# Driver API for library(distribution/hack)
#
# This module contains functions accessible from hack as defined
# by library(distribution/hack).  DON'T IMPORT this module directly;
# it will be imported by distribution/hack as needed.
#
# See IMPLEMENTING HACKS section of distribution/hack documentation
# for details.
#

hackr__name() {
    #
    # Print name of this hack
    #
    # Usage:
    #     hackr__name
    #
    test -n "$_hackr__hname" || {
        jat__log_error "hackr__name() called from outside hack handler"
        return 2
    }
    echo "$_hackr__hname"
}

hackr__action() {
    #
    # Print name of handled action
    #
    # Usage:
    #     hackr__action
    #
    test -n "$_hackr__aname" || {
        jat__log_error "hackr__action() called from outside hack handler"
        return 2
    }
    echo "$_hackr__aname"
}

hackr__util() {
    #
    # Call internal hack driver utility $1 with args $2..
    #
    # Usage:
    #     hackr__drvutil UTIL [ARG...]
    #
    # This function is intended to be called from within hack driver
    # and contains several utilities to help with application and cleanup
    # of common hacks such as AVC denial workarounds.
    #
    # Known UTILs are:
    #
    #  *  audit2hack: 'apply' action takes audit log on stdin and
    #     converts it to SELinux module, which it loads. 'cleanup'
    #     action just unloads said module.
    #
    # Note that utility "knows" which action is handled, so you don't
    # have to care, although ARGs might differ.
    #
    local UtilName=$1; shift    # UTIL name
    local util_args=("$@")      # ARGs
    local util_es               # exit status
    local util_fn               # name of implementing function
    local HackName              # current hack name
    local tmp                   # tempdir for util's needs
    util_fn="__hackr__util_$UtilName"
    test -n "$UtilName" || {
        jat__log_error "usage: hackr__drvutil UTIL"
        return 2
    }
    type -t "$util_fn" >/dev/null || {
        jat__log_error "unknown utility: $UtilName"
        return 2
    }
    case "$_hackr__aname" in
        apply|remove)  : ;;
        *) jat__log_error \
                "unknown action: $_hackr__aname"
            return 2 ;;
    esac
    HackName=$(hackr__name) || return 2
    tmp=$(mktemp -d -t "hackr__drvutil.$UtilName.XXXXXXXX")
    pushd "$tmp" >/dev/null || {
        jat__log_error "failed to cd: $tmp"
        return 3
    }
        ( "$util_fn" "${util_args[@]}" ); util_es=$?
    popd >/dev/null || {
        jat__log_error "could not popd from: $tmp"
        return 3
    }
    rm -rf "$tmp"
    return $util_es
}

#           #                                                            #
# TEMPLATES # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#           #                                                            #

hackr__util_SKEL() {
    #
    # Skeleton for hackr utility
    #
    # Note that utility will be called in subshell and in a temp
    # directory so you can drop files and change variables without
    # need to clean up.
    #
    case $_hackr__aname in
        apply)
            jat__cmd -h "$HackName: $UtilName: do something" \
              true
            ;;
        remove)
            jat__cmd -h "$HackName: $UtilName: revert something" \
              true
            ;;
    esac
}


#          #                      please do not apply hack to your brain #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                           that would let you past this line #


#
# Name of currently handled hack (set by d/hack)
#
_hackr__hname=

#
# Name of currently handled action (set by d/hack)
#
_hackr__aname=


__hackr__hw_isa_container() {
    #
    # True if this is a container
    #
    command -v systemd-detect-virt >/dev/null && {
        systemd-detect-virt -c >& /dev/null
        return $?
    }
    #shellcheck disable=SC2154
    test -n "$container"
}

__hackr__util_audit2hack() {
    #
    # Add/remove SELinux module built from AVC messages on stdin
    #
    __hackr__hw_isa_container && {
        jat__log_info "hackr__util audit2hack: this is container so nothing to do"
        return 0
    }
    case $_hackr__aname in
        apply)
            jat__eval -h "$HackName: $UtilName: build module" \
                "cat | audit2allow -M $HackName >/dev/null"
            jat__eval -h "$HackName: $UtilName: install module" \
                "semodule --install $HackName.pp"
            ;;
        remove)
            jat__eval -h "$HackName: $UtilName: remove module" \
                "semodule --remove $HackName"
            ;;
    esac
}

hackr__LibraryLoaded() {
    #
    # Do nothing (handler mandated by beakerlib)
    #
    true
}

#shellfu module-version=0.0.14
